/**
 * @fileoverview the `recommended-script` config for `eslint.config.js`
 * @author 唯然<weiran.zsd@outlook.com>
 * @deprecated use `flat/recommended-script` instead
 */

"use strict"

module.exports = require("../lib/index.js").configs["flat/recommended-script"]
