#!/usr/bin/env node
"use strict";const{getCompatibleVersions:e}=require("./index.cjs"),a=process.argv.slice(2),s={};for(let e=0;e<a.length;e++){const r=a[e];r&&("-h"===r||"--help"===r?s.help=!0:"--include-downstream-browsers"===r?s["include-downstream-browsers"]=!0:"--list-all-compatible-versions"===r?s["list-all-compatible-versions"]=!0:"--include-kaios"===r?s["include-kaios"]=!0:"--suppress-warnings"===r?s["suppress-warnings"]=!0:r.startsWith("--target-year=")?s["target-year"]=r.substring(14):"--target-year"===r?s["target-year"]=a[++e]:r.startsWith("--widely-available-on-date=")?s["widely-available-on-date"]=r.substring(27):"--widely-available-on-date"===r?s["widely-available-on-date"]=a[++e]:r.startsWith("--override-last-updated=")?s["override-last-updated"]=r.substring(24):"--override-last-updated"===r?s["override-last-updated"]=a[++e]:(console.error(`Unknown argument: ${r}`),process.exit(1)))}s.help&&(console.log("\nGet Baseline Widely available browser versions or Baseline year browser versions.\n\nUsage: baseline-browser-mapping [options]\n\nOptions:\n      --target-year                   Pass a year between 2015 and the current year to get browser versions compatible \n                                      with all Newly Available features as of the end of the year specified.\n      --widely-available-on-date      Pass a date in the format 'YYYY-MM-DD' to get versions compatible with Widely \n                                      available on the specified date.\n      --include-downstream-browsers   Whether to include browsers that use the same engines as a core Baseline browser.\n      --include-kaios                 Whether to include KaiOS in downstream browsers.  Requires --include-downstream-browsers.\n      --list-all-compatible-versions  Whether to include only the minimum compatible browser versions or all compatible versions.\n      --suppress-warnings             Supress potential warnings about data staleness when using a very recent feature cut off date.\n      --override-last-updated         Override the last updated date for the baseline data for debugging purposes.\n  -h, --help                          Show help\n\nExamples:\n  npx baseline-browser-mapping --target-year 2020\n  npx baseline-browser-mapping --widely-available-on-date 2023-04-05\n  npx baseline-browser-mapping --include-downstream-browsers\n  npx baseline-browser-mapping --list-all-compatible-versions\n".trim()),process.exit(0)),console.log(e({targetYear:s["target-year"]?Number.parseInt(s["target-year"]):void 0,widelyAvailableOnDate:s["widely-available-on-date"],includeDownstreamBrowsers:s["include-downstream-browsers"],listAllCompatibleVersions:s["list-all-compatible-versions"],includeKaiOS:s["include-kaios"],suppressWarnings:s["suppress-warnings"],overrideLastUpdated:s["override-last-updated"]?Number.parseInt(s["override-last-updated"]):void 0}));
