export type AliasedTags = {
    [key: string]: string[];
};
/**
 * @type {AliasedTags}
 */
export const closureTags: AliasedTags;
/**
 * @type {AliasedTags}
 */
export const jsdocTags: AliasedTags;
/**
 * @type {AliasedTags}
 */
export const typeScriptTags: AliasedTags;
