import { type ParsletFunction } from './Parslet.js';
export declare function createVariadicParslet({ allowPostfix, allowEnclosingBrackets }: {
    allowPostfix: boolean;
    allowEnclosingBrackets: boolean;
}): ParsletFunction;
