import type { Grammar } from './Grammar.js';
export declare const jsdocGrammar: Grammar;
export declare const jsdocNameGrammar: import("../parslets/Parslet.js").ParsletFunction[];
export declare const jsdocNamePathGrammar: import("../parslets/Parslet.js").ParsletFunction[];
export declare const jsdocNamePathSpecialGrammar: import("../parslets/Parslet.js").ParsletFunction[];
