export declare const closureGrammar: import("../parslets/Parslet.js").ParsletFunction[];
export declare const closureNameGrammar: import("../parslets/Parslet.js").ParsletFunction[];
export declare const closureNamePathGrammar: import("../parslets/Parslet.js").ParsletFunction[];
export declare const closureNamePathSpecialGrammar: import("../parslets/Parslet.js").ParsletFunction[];
