export {
	isLiteral,
	isStringLiteral,
	isNumericLiteral,
	isBigIntLiteral,
	isNullLiteral,
	isRegexLiteral,
	isEmptyStringLiteral,
} from './literal.js';

export {
	isNewExpression,
	isCallExpression,
	isCallOrNewExpression,
} from './call-or-new-expression.js';

export {default as isArrowFunctionBody} from './is-arrow-function-body.js';
export {default as isDirective} from './is-directive.js';
export {default as isEmptyNode} from './is-empty-node.js';
export {default as isEmptyArrayExpression} from './is-empty-array-expression.js';
export {default as isEmptyObjectExpression} from './is-empty-object-expression.js';
export {default as isExpressionStatement} from './is-expression-statement.js';
export {default as isFunction} from './is-function.js';
export {default as isMemberExpression} from './is-member-expression.js';
export {default as isMethodCall} from './is-method-call.js';
export {default as isNegativeOne} from './is-negative-one.js';
export {default as isReferenceIdentifier} from './is-reference-identifier.js';
export {default as isStaticRequire} from './is-static-require.js';
export {default as isTaggedTemplateLiteral} from './is-tagged-template-literal.js';
export {default as isUndefined} from './is-undefined.js';
export {default as functionTypes} from './function-types.js';
