import { Tokenizer } from './index.js';
/**
 * Splits the `@prefix` from remaining `Spec.lines[].token.description` into the `tag` token,
 * and populates `spec.tag`
 */
export default function tagTokenizer(): Tokenizer;
