import type { TransformRules } from './transform.js';
import type { NonRootResult } from '../result/NonRootResult.js';
export declare function identityTransformRules(): TransformRules<NonRootResult>;
