declare const _default: import("eslint").Rule.RuleModule;
export default _default;
export type ContextObject = {
    comment: string;
    context: string;
    message: string;
    minimum: import("../iterateJsdoc.js").Integer;
};
export type Context = string | ContextObject;
//# sourceMappingURL=noMissingSyntax.d.ts.map