/**
@import * as ESLint from 'eslint';
@import {UnicornCreate} from './to-eslint-create.js';
@import {UnicornRule} from './to-eslint-rule.js';
@import {UnicornContext} from './unicorn-context.js';
@import {TSESTree as Estree} from '@typescript-eslint/types';
*/

export {default as toEslintRules} from './to-eslint-rules.js';
