import * as eslint from 'eslint';

/**
 * Get ESLint class
 */
declare function getESLint(): typeof eslint.ESLint;
/**
 * Get LegacyESLint class
 */
declare function getLegacyESLint(): typeof eslint.ESLint;

export { getESLint, getLegacyESLint };
