/**
 * DON'T EDIT THIS FILE.
 * This file was generated by "scripts/update-lib-configs.js" script.
 */
"use strict"

module.exports = {
    extends: [
        require.resolve("./no-new-in-es2023-intl-api"),
        require.resolve("./no-new-in-es2022-intl-api"),
        require.resolve("./no-new-in-es2021-intl-api"),
        require.resolve("./no-new-in-es2020-intl-api"),
        require.resolve("./no-new-in-es2019-intl-api"),
        require.resolve("./no-new-in-es2018-intl-api"),
    ],
}
