/**
 * DON'T EDIT THIS FILE.
 * This file was generated by "scripts/update-lib-configs.js" script.
 */
"use strict"

module.exports = {
    plugins: ["es-x"],
    rules: {
        "es-x/no-intl-segmenter": "error",
        "es-x/no-intl-supportedvaluesof": "error",
    },
}
